@@ -7,12 +7,13 @@ import io
 import logging
 import os
 import pickle
+import warnings
 
 import numpy as np
 import zstandard as zstd
 
 
-__version__ = '0.1.1'
+__version__ = '0.2.0'
 
 
 @contextlib.contextmanager
@@ -32,8 +33,8 @@ def zstd_open_read(path, *args, **kwargs):
 
 
 class HashSink:
-    def __init__(self, h=hashlib.sha1()):
-        self._h = h
+    def __init__(self):
+        self._h = hashlib.sha1()
 
     def write(self, b: bytes):
         self._h.update(b)
@@ -46,83 +47,9 @@ class HashSink:
 
 
 def _hash(x):
-    logger = logging.getLogger(f'{__name__}._hash')
     sink = HashSink()
     pickle.dump(x, sink, protocol=4)
-    return sink.hexdigest()
-
-
-def _hash_args(*args, **kwargs):
-    logger = logging.getLogger(f'{__name__}._hash_args')
-    normalized_args = (args, tuple(sorted(kwargs.items())))
-    logger.debug('hash(args) = %s', _hash(normalized_args[0]))
-    logger.debug('hash(kwargs) = %s', _hash(normalized_args[1]))
-    return _hash(normalized_args)
-
-
-class Context:
-    def __init__(self):
-        self._product_dir = '.products'
-        self._modules = set()
-        self._producer = {}
-
-    def register_modules(self, *modules):
-        self._modules.update(modules)
-        for mod in modules:
-            for name in mod.output_names:
-                self._producer[name] = mod
-
-    def _get_path(self, name, module):
-        return os.path.join(
-            self._product_dir,
-            name,
-            module.version)
-
-    def has(self, name, module):
-        path = self._get_path(name, module)
-        return os.path.isfile(path + '.pickle.zst') or os.path.isfile(path + '.npz')
-
-    def get(self, name, module):
-        path = self._get_path(name, module)
-        if os.path.isfile(path + '.pickle.zst'):
-            with zstd_open_read(path + '.pickle.zst') as f:
-                return pickle.load(f)
-        elif os.path.isfile(path + '.npz'):
-            with np.load(path + '.npz') as npz:
-                if len(npz.files) == 1:
-                    return npz[npz.files[0]]
-                else:
-                    raise RuntimeError(f'Multiple files were found in npz: {path + ".npz"}')
-        else:
-            raise RuntimeError(f'Could not found data: "{name}"')
-
-    def put(self, name, module, data):
-        path = self._get_path(name, module)
-        os.makedirs(os.path.dirname(path), exist_ok=True)
-        if type(data) is np.ndarray:
-            np.savez_compressed(path + '.npz', data)
-        else:
-            with zstd_open_write(path + '.pickle.zst', level=19, threads=-1) as f:
-                pickle.dump(data, f, protocol=4)
-
-    def run(self, module):
-        if all(self.has(name, module) for name in module.output_names):
-            pass
-        else:
-            # Prepare input data for the module
-            for name in module.input_names:
-                producer = self._producer[name]
-                if not self.has(name, producer):
-                    # Recursively run the necessary modules
-                    self.run(producer)
-            # Execute the module
-            args = [self.get(name, self._producer[name]) for name in module.input_names]
-            products = module.execute(*args)
-            if len(module.output_names) == 1:
-                products = (products, )
-            # Store the products
-            for name, product in zip(module.output_names, products):
-                self.put(name, module, product)
+    return sink.digest()
 
 
 class Module(metaclass=ABCMeta):
@@ -130,6 +57,10 @@ class Module(metaclass=ABCMeta):
         self._input_names = input_names
         self._output_names = output_names
 
+    @property
+    def name(self):
+        return self.__class__.__name__
+
     @property
     def input_names(self):
         return self._input_names
@@ -139,8 +70,8 @@ class Module(metaclass=ABCMeta):
         return self._output_names
 
     @property
-    def version(self):
-        return hashlib.sha1(inspect.getsource(self.execute).encode('utf-8')).hexdigest()
+    def hash(self):
+        return hashlib.sha1(inspect.getsource(self.execute).encode('utf-8'))
 
     @abstractmethod
     def execute(self):
@@ -152,6 +83,9 @@ def module(input, output):
         class Wrapper(Module):
             def __init__(self):
                 super().__init__(input, output)
+            @property
+            def name(self):
+                return f.__name__
             def execute(self, *args, **kwargs):
                 return f(*args, **kwargs)
         wrapper = Wrapper()
@@ -159,28 +93,100 @@ def module(input, output):
     return decorator
 
 
-if __name__ == '__main__':
-    @module(input=[], output=['p1_output1', 'p1_output2', 'p1_output3'])
-    def process1():
-        """This is the docstring for process1."""
-        print('Executing process1')
-        return 'output1', 'output2', 'output3'
+class Context:
+    def __init__(self, product_dir='.products'):
+        self._product_dir = product_dir
+        self._modules = set()
+        self._producer = {}
 
-    class Process2(Module):
-        """This is the docstring for process2."""
-        def __init__(self):
-            super().__init__(
-                ['p1_output1', 'p1_output2', 'p1_output3'],
-                ['p2_output'])
+    def register_modules(self, *modules):
+        for mod in modules:
+            if not isinstance(mod, Module):
+                warnings.warn(f'Module "{mod.name}" is not an instance of Module', RuntimeWarning)
+            self._modules.add(mod)
+            for name in mod.output_names:
+                if name in self._producer:
+                    raise f'Producer for "{name}" is already registered'
+                self._producer[name] = mod
 
-        def execute(self, x, y, z):
-            print('Executing process2')
-            return x + y + z
+    def module(self, input, output):
+        decorator = module(input, output)
+        def deco(f):
+            module = decorator(f)
+            self.register_modules(module)
+            return module
+        return deco
 
-    process2 = Process2()
+    def _get_path(self, name, run_hash):
+        return os.path.join(
+            self._product_dir,
+            name,
+            run_hash)
 
-    ctx = Context()
-    ctx.register_modules(process1, process2)
-    ctx.run(process2)
+    def _has(self, name, run_hash):
+        path = self._get_path(name, run_hash)
+        return os.path.isfile(path + '.pickle.zst') or os.path.isfile(path + '.npz')
 
-    print(ctx.get('p2_output', process2))
+    def _get(self, name, run_hash, default=None):
+        path = self._get_path(name, run_hash)
+        if os.path.isfile(path + '.pickle.zst'):
+            with zstd_open_read(path + '.pickle.zst') as f:
+                return pickle.load(f)
+        elif os.path.isfile(path + '.npz'):
+            with np.load(path + '.npz') as npz:
+                if len(npz.files) == 1:
+                    return npz[npz.files[0]]
+                else:
+                    raise RuntimeError(f'Multiple files were found in npz: {path + ".npz"}')
+        else:
+            return default
+
+    def _put(self, name, run_hash, data):
+        path = self._get_path(name, run_hash)
+        os.makedirs(os.path.dirname(path), exist_ok=True)
+        if type(data) is np.ndarray:
+            np.savez_compressed(path + '.npz', data)
+        else:
+            with zstd_open_write(path + '.pickle.zst', level=19, threads=-1) as f:
+                pickle.dump(data, f, protocol=4)
+
+    def run(self, module, condition=None):
+        if condition is None:
+            condition = {}
+        # Run upstream modules
+        args = []
+        upstream_hashes = []
+        for name in module.input_names:
+            producer = self._producer[name]
+            i = producer.output_names.index(name)
+            upstream_products, upstream_hash = self.run(producer, condition)
+            args.append(upstream_products[i])
+            upstream_hashes.append(upstream_hash)
+        kwargs = condition.get(module.name, {})
+        # Compute a hash that identifies this run
+        h = hashlib.sha1()
+        h.update(hashlib.sha1(module.name.encode('utf-8')).digest())
+        h.update(module.hash.digest())
+        h.update(hashlib.sha1(''.join(upstream_hashes).encode('utf-8')).digest())
+        h.update(_hash(tuple(sorted(kwargs.items()))))
+        run_hash = h.hexdigest()
+        # Execute the module if needed
+        if all(self._has(name, run_hash) for name in module.output_names):
+            products = tuple(self._get(name, run_hash) for name in module.output_names)
+        else:
+            products = module.execute(*args, **kwargs)
+            if len(module.output_names) == 1:
+                products = (products, )
+            # Store the products
+            for name, product in zip(module.output_names, products):
+                self._put(name, run_hash, product)
+        # Return the result with the hash
+        return products, run_hash
+
+    def compute(self, name, condition=None):
+        if condition is None:
+            condition = {}
+        module = self._producer[name]
+        i = module.output_names.index(name)
+        products, _ = self.run(module, condition)
+        return products[i]
